from typing import TYPE_CHECKING, Any, Dict

from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class CodeCommit(Storage):
    """
    CodeCommit storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a CodeCommit repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has CodeCommit storage:

    ```python
    flow = Flow("my-flow")
    flow.storage = CodeCommit(repo="my/repo", path="/flows/flow.py")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register -f flow.py` to register this flow with CodeCommit storage.

    Args:
        - repo (str): the name of a CodeCommit repository to store this Flow
        - path (str, optional): a path pointing to a flow file in the repo
        - commit (str, optional): fully quaified reference that identifies the commit
          that contains the file. For example, you can specify a full commit ID, a tag,
          a branch name, or a reference such as refs/heads/master. If none is provided,
          the head commit is used
        - client_options (dict, optional): Additional options for the `boto3` client.
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        repo: str,
        path: str = None,
        commit: str = None,
        client_options: dict = None,
        **kwargs: Any
    ) -> None:
        self.flows = dict()  # type: Dict[str, str]
        self._flows = dict()  # type: Dict[str, "Flow"]
        self.repo = repo
        self.path = path
        self.commit = commit
        self.client_options = client_options

        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        client = self._boto3_client

        try:
            file_contents = client.get_file(
                repositoryName=self.repo,
                commitSpecifier=self.commit,
                filePath=flow_location,
            )["fileContent"]
            decoded_contents = file_contents.decode("utf-8")
        except Exception as exc:
            self.logger.error(
                "Error retrieving file contents from {} on repo {}. Ensure the file exists.".format(
                    flow_location, self.repo
                )
            )
            raise exc

        return extract_flow_from_file(
            file_contents=decoded_contents, flow_name=flow_name
        )

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    def build(self) -> "Storage":
        """
        Build the CodeCommit storage object and run basic healthchecks. Due to this object
        supporting file based storage no files are committed to the repository during
        this step. Instead, all files should be committed independently.

        Returns:
            - Storage: a CodeCommit object that contains information about how and where
                each flow is stored
        """
        self.run_basic_healthchecks()

        return self

    def __contains__(self, obj: Any) -> bool:
        """
        Method for determining whether an object is contained within this storage.
        """
        if not isinstance(obj, str):
            return False
        return obj in self.flows

    @property
    def _boto3_client(self):  # type: ignore
        from prefect.utilities.aws import get_boto_client

        kwargs = self.client_options or {}
        return get_boto_client(
            resource="codecommit", credentials=None, use_session=False, **kwargs
        )



import nbconvert
import nbformat
import papermill as pm

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class ExecuteNotebook(Task):
    """
    Task for running Jupyter Notebooks.
    In order to parametrize the notebook, you need to mark the parameters cell as described in
        the papermill documentation: https://papermill.readthedocs.io/en/latest/usage-parameterize.html

    Args:
        - path (string, optional): path to fetch the notebook from.
            Can be a cloud storage path.
            Can also be provided post-initialization by calling this task instance
        - parameters (dict, optional): dictionary of parameters to use for the notebook
            Can also be provided at runtime
        - output_format (str, optional): Notebook output format, should be a valid
            nbconvert Exporter name. 'json' is treated as 'notebook'.
            Valid exporter names: asciidoc, custom, html, latex, markdown,
            notebook, pdf, python, rst, script, slides, webpdf. (default: notebook)
        - exporter_kwargs (dict, optional): The arguments used for initializing
            the exporter.
        - kernel_name (string, optional): kernel name to run the notebook with.
            If not provided, the default kernel will be used.
        - **kwargs: additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        path: str = None,
        parameters: dict = None,
        output_format: str = "notebook",
        exporter_kwargs: dict = None,
        kernel_name: str = None,
        **kwargs
    ):
        self.path = path
        self.parameters = parameters
        self.output_format = output_format
        self.kernel_name = kernel_name
        self.exporter_kwargs = exporter_kwargs
        super().__init__(**kwargs)

    @defaults_from_attrs("path", "parameters", "output_format", "exporter_kwargs")
    def run(
        self,
        path: str = None,
        parameters: dict = None,
        output_format: str = None,
        exporter_kwargs: dict = None,
    ) -> str:
        """
        Run a Jupyter notebook and output as HTML, notebook, or other formats.

        Args:
        - path (string, optional): path to fetch the notebook from; can also be
            a cloud storage path
        - parameters (dict, optional): dictionary of parameters to use for the notebook
        - output_format (str, optional): Notebook output format, should be a valid
            nbconvert Exporter name. 'json' is treated as 'notebook'.
            Valid exporter names: asciidoc, custom, html, latex, markdown,
            notebook, pdf, python, rst, script, slides, webpdf. (default: notebook)
        - exporter_kwargs (dict, optional): The arguments used for initializing
            the exporter.
        """
        nb: nbformat.NotebookNode = pm.execute_notebook(
            path, "-", parameters=parameters, kernel_name=self.kernel_name
        )
        if output_format == "json":
            output_format = "notebook"

        if exporter_kwargs is None:
            exporter_kwargs = {}

        exporter = nbconvert.get_exporter(output_format)
        body, resources = nbconvert.export(exporter, nb, **exporter_kwargs)
        return body



from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from pyarrow import flight
import pyarrow as pa
from typing import Any, Dict


class HttpDremioClientAuthHandler(flight.ClientAuthHandler):
    """
    Client for Dremio authentication.

    Args:
        - username (str): user name used to authenticate
        - password (str): password used to authenticate

    Returns:
        - token (str)
    """

    def __init__(self, username: str, password: str):
        super(flight.ClientAuthHandler, self).__init__()
        self.basic_auth = flight.BasicAuth(username, password)
        self.token = None

    def authenticate(self, outgoing: Any, incoming: Any) -> None:
        auth = self.basic_auth.serialize()
        outgoing.write(auth)
        self.token = incoming.read()

    def get_token(self) -> str:
        return self.token


class DremioFetch(Task):
    """
    Task for fetching results of a query using Dremio Query Engine.

    Args:
        - user (str): user name used to authenticate
        - password (str): password used to authenticate
        - host (str): Dremio host address
        - port (int, optional): port used to connect to FlightClient, defaults to 32010 if not
            provided
        - query (str, optional): query to execute against Dremio
        - **kwargs (Any, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        user: str,
        password: str,
        host: str,
        port: int = 32010,
        query: str = None,
        **kwargs: Any,
    ):
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.query = query
        super().__init__(**kwargs)

    @defaults_from_attrs("user", "password", "host", "port", "query")
    def run(self, user: str, password: str, host: str, port: int, query: str) -> Dict:
        """
        Task run method. Executes a query against Dremio and fetches results.

        Args:
            - user (str): user name used to authenticate
            - password (str): password used to authenticate
            - host (str): Dremio host address
            - port (int, optional): port used to connect to FlightClient, defaults to 32010 if not
                provided
            - query (str, optional): query to execute against query engine

        Returns:
            - dict: a dictionary of data returned by Dremio

        Raises:
            - ValueError: if `query` is `None`
        """
        if not query:
            raise ValueError("A query string must be provided")

        client = flight.FlightClient(f"grpc+tcp://{host}:{port}")
        client.authenticate(HttpDremioClientAuthHandler(user, password))

        info = client.get_flight_info(
            flight.FlightDescriptor.for_command(query + "--arrow flight")
        )
        reader = client.do_get(info.endpoints[0].ticket)
        batches = []
        while True:
            try:
                batch, _ = reader.read_chunk()
                batches.append(batch)
            except StopIteration:
                break
        data = pa.Table.from_batches(batches)
        return data.to_pydict()



from marshmallow import fields

from prefect.utilities.serialization import (
    JSONCompatible,
    OneOfSchema,
    ObjectSchema,
    SortedList,
)
from prefect.run_configs import KubernetesRun, LocalRun, DockerRun, ECSRun, UniversalRun


class RunConfigSchemaBase(ObjectSchema):
    labels = SortedList(fields.String())


class UniversalRunSchema(RunConfigSchemaBase):
    class Meta:
        object_class = UniversalRun


class KubernetesRunSchema(RunConfigSchemaBase):
    class Meta:
        object_class = KubernetesRun

    job_template_path = fields.String(allow_none=True)
    job_template = JSONCompatible(allow_none=True)
    image = fields.String(allow_none=True)
    env = fields.Dict(keys=fields.String(), allow_none=True)
    cpu_limit = fields.String(allow_none=True)
    cpu_request = fields.String(allow_none=True)
    memory_limit = fields.String(allow_none=True)
    memory_request = fields.String(allow_none=True)
    service_account_name = fields.String(allow_none=True)
    image_pull_secrets = fields.List(fields.String(), allow_none=True)


class ECSRunSchema(RunConfigSchemaBase):
    class Meta:
        object_class = ECSRun

    task_definition_path = fields.String(allow_none=True)
    task_definition = JSONCompatible(allow_none=True)
    task_definition_arn = fields.String(allow_none=True)
    image = fields.String(allow_none=True)
    env = fields.Dict(keys=fields.String(), allow_none=True)
    cpu = fields.String(allow_none=True)
    memory = fields.String(allow_none=True)
    task_role_arn = fields.String(allow_none=True)
    execution_role_arn = fields.String(allow_none=True)
    run_task_kwargs = JSONCompatible(allow_none=True)


class LocalRunSchema(RunConfigSchemaBase):
    class Meta:
        object_class = LocalRun

    env = fields.Dict(keys=fields.String(), allow_none=True)
    working_dir = fields.String(allow_none=True)


class DockerRunSchema(RunConfigSchemaBase):
    class Meta:
        object_class = DockerRun

    image = fields.String(allow_none=True)
    env = fields.Dict(keys=fields.String(), allow_none=True)


class RunConfigSchema(OneOfSchema):
    type_schemas = {
        "KubernetesRun": KubernetesRunSchema,
        "ECSRun": ECSRunSchema,
        "LocalRun": LocalRunSchema,
        "DockerRun": DockerRunSchema,
        "UniversalRun": UniversalRunSchema,
    }



import gzip
import io
import uuid
import pendulum

from prefect import Task
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class S3Download(Task):
    """
    Task for downloading data from an S3 bucket and returning it as a string.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set the `AWS_CREDENTIALS` Prefect Secret
    containing your AWS access keys which will be passed directly to the `boto3` client, or you
    can [configure your flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - bucket (str, optional): the name of the S3 Bucket to download from
        - boto_kwargs (dict, optional): additional keyword arguments to forward to the boto client.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(self, bucket: str = None, boto_kwargs: dict = None, **kwargs):
        self.bucket = bucket

        if boto_kwargs is None:
            self.boto_kwargs = {}
        else:
            self.boto_kwargs = boto_kwargs

        super().__init__(**kwargs)

    @defaults_from_attrs("bucket")
    def run(
        self,
        key: str,
        credentials: str = None,
        bucket: str = None,
        compression: str = None,
    ):
        """
        Task run method.

        Args:
            - key (str): the name of the Key within this bucket to retrieve
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
            - bucket (str, optional): the name of the S3 Bucket to download from
            - compression (str, optional): specifies a file format for decompression, decompressing
                data upon download. Currently supports `'gzip'`.

        Returns:
            - str: the contents of this Key / Bucket, as a string
        """
        if bucket is None:
            raise ValueError("A bucket name must be provided.")

        s3_client = get_boto_client(
            "s3", credentials=credentials, use_session=True, **self.boto_kwargs
        )

        stream = io.BytesIO()

        # download
        s3_client.download_fileobj(Bucket=bucket, Key=key, Fileobj=stream)

        # prepare data and return
        stream.seek(0)
        output = stream.read()

        # decompress data if decompression is specified
        if compression:
            if compression == "gzip":
                output = gzip.decompress(output)
            else:
                raise ValueError(f"Unrecognized compression method '{compression}'.")

        return output.decode()


class S3Upload(Task):
    """
    Task for uploading string data (e.g., a JSON string) to an S3 bucket.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set a Prefect Secret containing your AWS
    access keys which will be passed directly to the `boto3` client, or you can [configure your
    flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - bucket (str, optional): the name of the S3 Bucket to upload to
        - boto_kwargs (dict, optional): additional keyword arguments to forward to the boto client.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(self, bucket: str = None, boto_kwargs: dict = None, **kwargs):
        self.bucket = bucket

        if boto_kwargs is None:
            self.boto_kwargs = {}
        else:
            self.boto_kwargs = boto_kwargs

        super().__init__(**kwargs)

    @defaults_from_attrs("bucket")
    def run(
        self,
        data: str,
        key: str = None,
        credentials: dict = None,
        bucket: str = None,
        compression: str = None,
    ):
        """
        Task run method.

        Args:
            - data (str): the data payload to upload
            - key (str, optional): the Key to upload the data under; if not
                provided, a random `uuid` will be created
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
            - bucket (str, optional): the name of the S3 Bucket to upload to
            - compression (str, optional): specifies a file format for compression,
                compressing data before upload. Currently supports `'gzip'`.

        Returns:
            - str: the name of the Key the data payload was uploaded to
        """
        if bucket is None:
            raise ValueError("A bucket name must be provided.")

        s3_client = get_boto_client(
            "s3", credentials=credentials, use_session=True, **self.boto_kwargs
        )

        # compress data if compression is specified
        if compression:
            if compression == "gzip":
                data = gzip.compress(data)
            else:
                raise ValueError(f"Unrecognized compression method '{compression}'.")

        # prepare data
        try:
            stream = io.BytesIO(data)
        except TypeError:
            stream = io.BytesIO(data.encode())

        # create key if not provided
        if key is None:
            key = str(uuid.uuid4())

        # upload
        s3_client.upload_fileobj(stream, Bucket=bucket, Key=key)
        return key


class S3List(Task):
    """
    Task for listing files from an S3 bucket.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    For authentication, there are two options: you can set the `AWS_CREDENTIALS` Prefect Secret
    containing your AWS access keys which will be passed directly to the `boto3` client, or you
    can [configure your flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - bucket (str, optional): the name of the S3 Bucket to list the files of.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(self, bucket: str = None, **kwargs):
        self.bucket = bucket
        super().__init__(**kwargs)

    @defaults_from_attrs("bucket")
    def run(
        self,
        prefix: str,
        delimiter: str = "",
        page_size: int = None,
        max_items: int = None,
        credentials: str = None,
        bucket: str = None,
        last_modified_begin: str = None,
        last_modified_end: str = None,
    ):
        """
        Task run method.

        Args:
            - prefix (str): the name of the prefix within this bucket to retrieve objects from
            - delimiter (str): indicates the key hierarchy
            - page_size (int): controls the number of items returned per page of each result
            - max_items (int): limits the maximum number of total items returned during pagination
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
            - bucket (str, optional): the name of the S3 Bucket to list the files of
            - last_modified_begin (str, optional): keep items with `LastModified` greater than
                or equal to given value. timestamp should be in `RFC 3339`, `ISO 8601` or any
                pendulum supported [format](https://pendulum.eustace.io/docs/#parsing).
            - last_modified_end (str, optional): keep items with `LastModified` less than
                or equal to given value. timestamp should be in `RFC 3339`, `ISO 8601` or any
                pendulum supported [format](https://pendulum.eustace.io/docs/#parsing).

        Returns:
            - list[str]: A list of keys that match the given prefix.
        """
        if bucket is None:
            raise ValueError("A bucket name must be provided.")

        s3_client = get_boto_client("s3", credentials=credentials)

        config = {"PageSize": page_size, "MaxItems": max_items}
        paginator = s3_client.get_paginator("list_objects_v2")
        results = paginator.paginate(
            Bucket=bucket, Prefix=prefix, Delimiter=delimiter, PaginationConfig=config
        )

        filters = []
        filtered_results = None
        # create the parts of JMESPath query
        if last_modified_end:
            filters.append(
                "(to_string(LastModified) <= '\"{}\"')".format(
                    pendulum.parse(last_modified_end).to_datetime_string()
                )
            )
        if last_modified_begin:
            filters.append(
                "(to_string(LastModified) >= '\"{}\"')".format(
                    pendulum.parse(last_modified_begin).to_datetime_string()
                )
            )

        if filters:
            # combine the parts and create the complete JMESPath query
            filtered_results = results.search(
                '{{"Contents": Contents[? {} ]}}'.format(" && ".join(filters))
            )

        files = []
        for page in filtered_results if filtered_results else results:
            files.extend(obj["Key"] for obj in page.get("Contents", []))

        return files



import requests
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs
from typing import Any

import json


class CreateItem(Task):
    """
    Task for creating items in a Monday board

    Args:
        - board_id (int): the id of the board to add the new item
        - group_id (str): the id of the group to add the new item
        - item_name (str): the name of the item to be created
        - column_values (dict, optional): any additional custom columns added to your board
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        board_id: int = None,
        group_id: str = None,
        item_name: str = None,
        column_values: dict = None,
        **kwargs: Any
    ):
        self.board_id = board_id
        self.group_id = group_id
        self.item_name = item_name
        self.column_values = column_values

        super().__init__(**kwargs)

    @defaults_from_attrs("board_id", "group_id", "item_name", "column_values")
    def run(
        self,
        board_id: str = None,
        group_id: str = None,
        item_name: str = None,
        column_values: dict = None,
        monday_api_token: str = "MONDAY_API_TOKEN",
    ) -> None:
        """
        Task run method.

        Args:
            - board_id (int): the id of the board to add the new item
            - group_id (str): the id of the group to add the new item
            - item_name (str): the name of the item to be created
            - column_values (dict, optional): any additional custom columns added to your board
            - monday_api_token (str): the name of the Prefect Secret which stored your Monday
                API Token.

        Returns:
            - int: the id of the item created
        """

        if board_id is None:
            raise ValueError("A board id must be provided.")

        if group_id is None:
            raise ValueError("A group id must be provided")

        if item_name is None:
            raise ValueError("An item name must be provided")

        if monday_api_token is None:
            raise ValueError("An API Token must be provided")

        # prepare mutation
        headers = {"Authorization": monday_api_token}

        variables = {
            "MONDAY_BOARD_ID": board_id,
            "MONDAY_GROUP_ID": group_id,
            "ITEM_NAME": item_name,
        }
        if column_values:
            variables.update({"COLUMN_VALUES": json.dumps(column_values)})

        query = """
            mutation (
                $MONDAY_BOARD_ID:Int!,
                $MONDAY_GROUP_ID:String!,
                $ITEM_NAME:String!,
                $COLUMN_VALUES:JSON) {
                create_item (
                    board_id: $MONDAY_BOARD_ID,
                    group_id: $MONDAY_GROUP_ID,
                    item_name: $ITEM_NAME,
                    column_values: $COLUMN_VALUES)
                {
                    id
                }
            }
        """

        response = requests.post(
            "https://api.monday.com/v2/",
            json={"query": query, "variables": variables},
            headers=headers,
        )
        if response.status_code == 200:
            return response.json()["data"]["create_item"]["id"]
        else:
            raise Exception("Query failed {}. {}".format(response.status_code, query))



import click

import prefect
from prefect.client import Client
from prefect.engine import get_default_flow_runner_class
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities.graphql import with_args


@click.group(hidden=True)
def execute():
    """
    Execute flow runs.

    \b
    Usage:
        $ prefect execute [OBJECT]

    \b
    Arguments:
        flow-run  Execute a flow run with a backend API

    \b
    Examples:
        $ prefect execute flow-run
    """


@execute.command(hidden=True)
def flow_run():
    """
    Execute a flow run in the context of a backend API.
    """
    flow_run_id = prefect.context.get("flow_run_id")
    if not flow_run_id:
        click.echo("Not currently executing a flow within a Cloud context.")
        raise Exception("Not currently executing a flow within a Cloud context.")

    query = {
        "query": {
            with_args("flow_run", {"where": {"id": {"_eq": flow_run_id}}}): {
                "flow": {"name": True, "storage": True, "run_config": True},
                "version": True,
            }
        }
    }

    client = Client()
    result = client.graphql(query)
    flow_run = result.data.flow_run

    if not flow_run:
        click.echo("Flow run {} not found".format(flow_run_id))
        raise ValueError("Flow run {} not found".format(flow_run_id))

    try:
        flow_data = flow_run[0].flow
        storage_schema = prefect.serialization.storage.StorageSchema()
        storage = storage_schema.load(flow_data.storage)

        # populate global secrets
        secrets = prefect.context.get("secrets", {})
        for secret in storage.secrets:
            secrets[secret] = PrefectSecret(name=secret).run()

        with prefect.context(secrets=secrets, loading_flow=True):
            flow = storage.get_flow(flow_data.name)

        with prefect.context(secrets=secrets):
            if flow_data.run_config is not None:
                runner_cls = get_default_flow_runner_class()
                runner_cls(flow=flow).run()
            else:
                environment = flow.environment
                environment.setup(flow)
                environment.execute(flow)
    except Exception as exc:
        msg = "Failed to load and execute Flow's environment: {}".format(repr(exc))
        state = prefect.engine.state.Failed(message=msg)
        client.set_flow_run_state(flow_run_id=flow_run_id, state=state)
        client.write_run_logs(
            dict(
                flow_run_id=flow_run_id,  # type: ignore
                name="execute flow-run",
                message=msg,
                level="ERROR",
            )
        )
        click.echo(str(exc))
        raise exc



import warnings
from typing import Any

from prefect.executors import LocalDaskExecutor as _LocalDaskExecutor
from prefect.executors import DaskExecutor as _DaskExecutor


class DaskExecutor(_DaskExecutor):
    def __new__(cls, *args: Any, **kwargs: Any) -> "DaskExecutor":
        warnings.warn(
            "prefect.engine.executors.DaskExecutor has been moved to "
            "`prefect.executors.DaskExecutor`, please update your imports",
            stacklevel=2,
        )
        return super().__new__(cls)


class LocalDaskExecutor(_LocalDaskExecutor):
    def __new__(cls, *args: Any, **kwargs: Any) -> "LocalDaskExecutor":
        warnings.warn(
            "prefect.engine.executors.LocalDaskExecutor has been moved to "
            "`prefect.executors.LocalDaskExecutor`, please update your imports",
            stacklevel=2,
        )
        return super().__new__(cls)



"""
This module contains a collection of tasks for interacting with SQL Server databases via
the pyodbc library.
"""

try:
    from prefect.tasks.sql_server.sql_server import (
        SqlServerExecute,
        SqlServerExecuteMany,
        SqlServerFetch,
    )
except ImportError:
    raise ImportError(
        'Using `prefect.tasks.sql_server` requires Prefect to be installed with the "sql_server" extra.'
    )



"""
Utility functions for interacting with git-based clients.
"""
import os
import prefect
import requests

from typing import Any

try:
    from github import Github
except ImportError:
    Github = None  # type: ignore

try:
    from gitlab import Gitlab
except ImportError:
    Gitlab = None  # type: ignore

try:
    from atlassian import Bitbucket
except ImportError:
    Bitbucket = None  # type: ignore


def get_github_client(credentials: dict = None, **kwargs: Any) -> "Github":
    """
    Utility function for loading github client objects from a given set of credentials.

    Args:
        - credentials (dict, optional): a dictionary of Github credentials used to
            initialize the Client; if not provided, will attempt to load the
            Client using ambient environment settings
        - **kwargs (Any, optional): additional keyword arguments to pass to the github Client

    Returns:
        - Client: an initialized and authenticated github Client
    """
    if not Github:
        raise ImportError(
            "Unable to import Github, please ensure you have installed the github extra"
        )

    access_token = None

    if credentials:
        access_token = credentials.get("GITHUB_ACCESS_TOKEN")
    else:
        access_token = prefect.context.get("secrets", {}).get(
            "GITHUB_ACCESS_TOKEN", None
        )

    # Attempt to grab out of env if not provided directly or through Prefect Secret
    if not access_token:
        access_token = os.getenv("GITHUB_ACCESS_TOKEN", None)

    return Github(access_token, **kwargs)


def get_gitlab_client(
    credentials: dict = None, host: str = None, **kwargs: Any
) -> "Gitlab":
    """
    Utility function for loading gitlab client objects from a given set of credentials.

    Args:
        - credentials (dict, optional): a dictionary of Gitlab credentials used to
            initialize the Client; if not provided, will attempt to load the
            Client using ambient environment settings
        - host (str, optional): the host string for gitlab server users. If not provided, defaults
            to https://gitlab.com
        - **kwargs (Any, optional): additional keyword arguments to pass to the gitlab Client

    Returns:
        - Client: an initialized and authenticated gitlab Client
    """
    if not Gitlab:
        raise ImportError(
            "Unable to import Gitlab, please ensure you have installed the gitlab extra"
        )

    if credentials:
        access_token = credentials.get("GITLAB_ACCESS_TOKEN")
    else:
        access_token = prefect.context.get("secrets", {}).get(
            "GITLAB_ACCESS_TOKEN", None
        )

    if not access_token:
        access_token = os.getenv("GITLAB_ACCESS_TOKEN", None)

    if not host:
        host = "https://gitlab.com"

    return Gitlab(host, private_token=access_token, **kwargs)


def get_bitbucket_client(
    credentials: dict = None, host: str = None, **kwargs: Any
) -> "Bitbucket":
    """
    Utility function for loading Bitbucket client objects from a given set of credentials.

    Args:
        - credentials (dict, optional): a dictionary of Bitbucket credentials used to
            initialize the Client; if not provided, will attempt to load the
            Client using ambient environment settings
        - host (str, optional): the host string for bitbucket server users. If not provided, defaults
            to https://bitbucket.org
        - **kwargs (Any, optional): additional keyword arguments to pass to the Bitbucket Client
            Bitbucket accepts: "cloud", "api_version", "api_root"

    Returns:
        - Client: an initialized and authenticated Bitbucket Client
    """
    if not Bitbucket:
        raise ImportError(
            "Unable to import Bitbucket, please ensure you have installed the bitbucket extra"
        )

    if credentials:
        access_token = credentials["BITBUCKET_ACCESS_TOKEN"]
    else:
        access_token = prefect.context.get("secrets", {}).get(
            "BITBUCKET_ACCESS_TOKEN", None
        )

    if not access_token:
        access_token = os.getenv("BITBUCKET_ACCESS_TOKEN", None)

    if not host:
        host = "https://bitbucket.org"

    session = requests.Session()
    if access_token is None:
        session.headers["Authorization"] = "Bearer "
    else:
        session.headers["Authorization"] = "Bearer " + access_token
    return Bitbucket(host, session=session, **kwargs)



import io
import os
from urllib.parse import urlparse
from typing import NamedTuple


class ParsedPath(NamedTuple):
    """A parsed path"""

    scheme: str
    netloc: str
    path: str


def parse_path(path: str) -> ParsedPath:
    """Parse a path into its components.

    Parses a path of the form `[{scheme}://][{netloc}]{path}`, similar to
    `urllib.parse.urlparse`. The main difference is that windows paths with
    `\\` and optional drive designators are supported.

    Args:
        - path (str): The path to parse.

    Returns:
        - ParsedPath: the parsed path.
    """
    # On windows, this might be a local path with a drive
    # splitdrive is a no-op on non-windows
    drive, _ = os.path.splitdrive(path)
    if drive:
        scheme = "file"
        netloc = ""
    else:
        parsed = urlparse(path)
        scheme = parsed.scheme or ""
        netloc = parsed.netloc or ""
        path = parsed.path or ""

    if scheme in (None, "", "local"):
        scheme = "file"

    return ParsedPath(scheme, netloc, path)


def read_bytes_from_path(path: str) -> bytes:
    """Read bytes from a given path.

    Paths may be local files, or remote files (given a supported file scheme).

    Args:
        - path (str): The file path

    Returns:
        - bytes: The file contents
    """
    parsed = parse_path(path)
    if not parsed.scheme or parsed.scheme in ("file", "agent"):
        with open(parsed.path, "rb") as f:
            return f.read()
    elif parsed.scheme == "gcs":
        from prefect.utilities.gcp import get_storage_client

        client = get_storage_client()
        bucket = client.bucket(parsed.netloc)
        blob = bucket.get_blob(parsed.path.lstrip("/"))
        if blob is None:
            raise ValueError(f"Job template doesn't exist at {path}")
        # Support GCS < 1.31
        return (
            blob.download_as_bytes()
            if hasattr(blob, "download_as_bytes")
            else blob.download_as_string()
        )
    elif parsed.scheme == "s3":
        from prefect.utilities.aws import get_boto_client

        client = get_boto_client(resource="s3")
        stream = io.BytesIO()
        client.download_fileobj(
            Bucket=parsed.netloc, Key=parsed.path.lstrip("/"), Fileobj=stream
        )
        return stream.getvalue()
    else:
        raise ValueError(f"Unsupported file scheme {path}")



import warnings
from typing import Any

import prefect
from prefect import Task
from prefect.client import Client
from prefect.utilities.tasks import defaults_from_attrs


class CancelFlowRun(Task):
    """
    Task to cancel a flow run. If `flow_run_id` is not provided,
    `flow_run_id` from `prefect.context` will be used by default

    Args:
        - flow_run_id (str, optional): The ID of the flow run to cancel
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        flow_run_id: str = None,
        **kwargs: Any,
    ):
        self.flow_run_id = flow_run_id
        super().__init__(**kwargs)

    @defaults_from_attrs("flow_run_id")
    def run(self, flow_run_id: str = None) -> bool:
        """
        Args:
            - flow_run_id (str, optional): The ID of the flow run to cancel

        Returns:
            - bool: Whether the flow run was canceled successfully or not
        """
        flow_run_id = flow_run_id or prefect.context.get("flow_run_id")
        if not flow_run_id:
            raise ValueError("Can't cancel a flow run without flow run ID.")

        client = Client()
        return client.cancel_flow_run(flow_run_id)


class CancelFlowRunTask(CancelFlowRun):
    def __new__(cls, *args, **kwargs):  # type: ignore
        warnings.warn(
            "`CancelFlowRunTask` has been renamed to `prefect.tasks.prefect.CancelFlowRun`,"
            "please update your code accordingly",
            stacklevel=2,
        )
        return super().__new__(cls)



"""
Execution environments encapsulate the logic for where your Flow should execute in Prefect Cloud.

DEPRECATED: Environment based configuration is deprecated, please transition to
configuring `flow.run_config` instead of `flow.environment`. See
https://docs.prefect.io/orchestration/flow_config/overview.html for more info.
"""
from prefect.environments.execution.base import Environment, load_and_run_flow
from prefect.environments.execution.dask import DaskKubernetesEnvironment
from prefect.environments.execution.dask import DaskCloudProviderEnvironment
from prefect.environments.execution.fargate import FargateTaskEnvironment
from prefect.environments.execution.k8s import KubernetesJobEnvironment
from prefect.environments.execution.local import LocalEnvironment



from distutils.version import LooseVersion

from prefect.utilities.graphql import GraphQLResult


def get_flow_image(flow_run: GraphQLResult) -> str:
    """
    Retrieve the image to use for this flow run deployment.

    Args:
        - flow_run (GraphQLResult): A GraphQLResult flow run object

    Returns:
        - str: a full image name to use for this flow run

    Raises:
        - ValueError: if deployment attempted on unsupported Storage type and `image` not
            present in environment metadata
    """
    from prefect.storage import Docker
    from prefect.serialization.storage import StorageSchema
    from prefect.serialization.run_config import RunConfigSchema
    from prefect.serialization.environment import EnvironmentSchema

    has_run_config = getattr(flow_run, "run_config", None) is not None
    has_environment = getattr(flow_run.flow, "environment", None) is not None

    storage = StorageSchema().load(flow_run.flow.storage)
    # Not having an environment implies run-config based flow, even if
    # run_config is None.
    if has_run_config or not has_environment:
        if isinstance(storage, Docker):
            return storage.name
        elif has_run_config:
            run_config = RunConfigSchema().load(flow_run.run_config)
            if getattr(run_config, "image", None) is not None:
                return run_config.image
        # No image found on run-config, and no environment present. Use default.
        # core_version should always be present, but just in case
        version = flow_run.flow.get("core_version") or "latest"
        cleaned_version = version.split("+")[0]
        return f"prefecthq/prefect:{cleaned_version}"
    else:
        environment = EnvironmentSchema().load(flow_run.flow.environment)
        if hasattr(environment, "metadata") and hasattr(environment.metadata, "image"):
            return environment.metadata.get("image")
        elif isinstance(storage, Docker):
            return storage.name
        raise ValueError(
            f"Storage for flow run {flow_run.id} is not of type Docker and "
            f"environment has no `image` attribute in the metadata field."
        )


def get_flow_run_command(flow_run: GraphQLResult) -> str:
    """
    Determine the flow run command to use based on a flow's version. This is due to a command
    deprecation in `0.13.0` where `execute cloud-flow` was changes to `execute flow-run`

    Args:
        - flow_run (GraphQLResult): A GraphQLResult flow run object

    Returns:
        - str: a prefect CLI command to execute a flow run
    """
    core_version = getattr(flow_run.flow, "core_version", None) or "0.0.0"

    if LooseVersion(core_version) < LooseVersion("0.13.0"):
        return "prefect execute cloud-flow"

    return "prefect execute flow-run"



"""
Utility functions for interacting with AWS.
"""
import prefect

import boto3
from typing import Any, Optional


def get_boto_client(
    resource: str,
    credentials: Optional[dict] = None,
    use_session: Optional[bool] = False,
    profile_name: Optional[str] = None,
    **kwargs: Any
) -> "boto3.client":
    """
    Utility function for loading boto3 client objects from a given set of credentials.

    Args:
        - resource (str): the name of the resource to retrieve a client for
        - credentials (dict, optional): a dictionary of AWS credentials used to
            initialize the Client; if not provided, will attempt to load the
            Client using ambient environment settings
        - use_session (bool, optional): a boolean specifying whether to load
            this client using a session or not; defaults to `False`
        - profile_name (str, optional): The name of a profile to use.
        - **kwargs (Any, optional): additional keyword arguments to pass to boto3

    Returns:
        - Client: an initialized and authenticated boto3 Client
    """

    if profile_name and not use_session:
        raise ValueError(
            "profile_name can only be used with the boto3.session. Please set use_session=True"
        )

    if credentials:
        aws_access_key = credentials["ACCESS_KEY"]
        aws_secret_access_key = credentials["SECRET_ACCESS_KEY"]
        aws_session_token = credentials.get("SESSION_TOKEN")
    else:
        ctx_credentials = prefect.context.get("secrets", {}).get("AWS_CREDENTIALS", {})
        aws_access_key = ctx_credentials.get("ACCESS_KEY")
        aws_secret_access_key = ctx_credentials.get("SECRET_ACCESS_KEY")
        aws_session_token = ctx_credentials.get("SESSION_TOKEN")

    kwargs_access_key_id = kwargs.pop("aws_access_key_id", None)
    kwargs_secret_access_key = kwargs.pop("aws_secret_access_key", None)
    kwargs_session_token = kwargs.pop("aws_session_token", None)

    aws_access_key = aws_access_key or kwargs_access_key_id
    aws_secret_access_key = aws_secret_access_key or kwargs_secret_access_key
    aws_session_token = aws_session_token or kwargs_session_token

    if use_session:
        # see https://boto3.amazonaws.com/v1/documentation/api/latest/guide/resources.html?#multithreading-multiprocessing  # noqa
        region_name = kwargs.pop("region_name", None)
        botocore_session = kwargs.pop("botocore_session", None)
        session = boto3.session.Session(
            profile_name=profile_name,
            region_name=region_name,
            botocore_session=botocore_session,
        )
        return session.client(
            resource,
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            **kwargs
        )
    else:
        return boto3.client(
            resource,
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            **kwargs
        )



import binascii
import importlib
import json
import sys
import warnings
from operator import attrgetter
from typing import TYPE_CHECKING
from distutils.version import LooseVersion

import cloudpickle

import prefect
from prefect.utilities.exceptions import StorageError

if TYPE_CHECKING:
    from prefect.core.flow import Flow  # pylint: disable=W0611


def get_flow_image(flow: "Flow") -> str:
    """
    Retrieve the image to use for this flow deployment. Will start by looking for
    an `image` value in the flow's `environment.metadata`. If not found then it will fall
    back to using the `flow.storage`.

    Args:
        - flow (Flow): A flow object

    Returns:
        - str: a full image name to use for this flow run

    Raises:
        - ValueError: if deployment attempted on unsupported Storage type and `image` not
            present in environment metadata
    """
    environment = flow.environment
    if (
        environment is not None
        and hasattr(environment, "metadata")
        and environment.metadata.get("image")
    ):
        return environment.metadata.get("image", "")
    else:
        storage = flow.storage
        if not isinstance(storage, prefect.storage.Docker):
            raise ValueError(
                f"Storage for flow run {flow.name} is not of type Docker and "
                f"environment has no `image` attribute in the metadata field."
            )

        return storage.name


def extract_flow_from_file(
    file_path: str = None, file_contents: str = None, flow_name: str = None
) -> "Flow":
    """
    Extract a flow object from a file.

    Args:
        - file_path (str, optional): A file path pointing to a .py file containing a flow
        - file_contents (str, optional): The string contents of a .py file containing a flow
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not set then the first flow object retrieved from file will be returned.

    Returns:
        - Flow: A flow object extracted from a file

    Raises:
        - ValueError: if both `file_path` and `file_contents` are provided or neither are.
    """
    if file_path and file_contents:
        raise ValueError("Provide either `file_path` or `file_contents` but not both.")

    if not file_path and not file_contents:
        raise ValueError("Provide either `file_path` or `file_contents`.")

    # Read file contents
    if file_path:
        with open(file_path, "r") as f:
            contents = f.read()

    # Use contents directly
    if file_contents:
        contents = file_contents

    # Load objects from file into dict
    exec_vals = {}  # type: ignore
    exec(contents, exec_vals)

    # Grab flow name from values loaded via exec
    for var in exec_vals:
        if isinstance(exec_vals[var], prefect.Flow):
            if flow_name and exec_vals[var].name == flow_name:
                return exec_vals[var]
            elif not flow_name:
                return exec_vals[var]

    raise ValueError("No flow found in file.")


def extract_flow_from_module(module_str: str, flow_name: str = None) -> "Flow":
    """
    Extract a flow object from a python module.

    Args:
        - module_str (str): A module path pointing to a .py file containing a flow.
            For example, 'myrepo.mymodule.myflow' where myflow.py contains the flow.
            Additionally, `:` can be used to access module's attribute, for example,
            'myrepo.mymodule.myflow:flow' or 'myrepo.mymodule.myflow:MyObj.newflow'.
        - flow_name (str, optional): A specific name of a flow to extract from a file.
            If not set then the first flow object retrieved from file will be returned.
            The "first" flow object will be based on the dir(module) ordering, which
            is alphabetical and capitalized letters come first.

    Returns:
        - Flow: A flow object extracted from a file
    """

    # load the module
    module_parts = module_str.split(":", 2)

    if len(module_parts) == 2 and flow_name is not None:
        raise ValueError(
            "Provide either `module_str` without an attribute specifier or remove `flow_name`."
        )
    elif len(module_parts) == 2:
        module_name, flow_name = module_parts
    else:
        module_name = module_str

    module = importlib.import_module(module_name)

    # if flow_name is specified, grab it from the module
    if flow_name:
        attr = attrgetter(flow_name)(module)

        if callable(attr):
            attr = attr()

        if isinstance(attr, prefect.Flow):
            return attr
        else:
            raise ValueError(
                f"{module_name}:{flow_name} must return `prefect.Flow`, not {type(attr)}."
            )
    # otherwise loop until we get a Flow object
    else:
        for var in dir(module):
            attr = getattr(module, var)
            if isinstance(attr, prefect.Flow):
                return attr

    raise ValueError("No flow found in module.")


def flow_to_bytes_pickle(flow: "Flow") -> bytes:
    """Serialize a flow to bytes.

    The flow is serialized using `cloudpickle`, with some extra metadata on
    included via JSON. The flow can be reloaded using `flow_from_bytes_pickle`.

    Args:
        - flow (Flow): the flow to be serialized.

    Returns:
        - bytes: a serialized representation of the flow.
    """
    flow_data = binascii.b2a_base64(
        cloudpickle.dumps(flow, protocol=4), newline=False
    ).decode("utf-8")
    out = json.dumps({"flow": flow_data, "versions": _get_versions()})
    return out.encode("utf-8")


def _get_versions() -> dict:
    """Get version info on libraries where a version-mismatch between
    registration and execution environment may cause a flow to fail to load
    properly"""
    return {
        "cloudpickle": cloudpickle.__version__,
        "prefect": prefect.__version__,
        "python": "%d.%d.%d" % sys.version_info[:3],
    }


def flow_from_bytes_pickle(data: bytes) -> "Flow":
    """Load a flow from bytes."""
    try:
        info = json.loads(data.decode("utf-8"))
    except Exception:
        # Serialized using older version of prefect, use cloudpickle directly
        flow_bytes = data
        reg_versions = {}
    else:
        flow_bytes = binascii.a2b_base64(info["flow"])
        reg_versions = info["versions"]

    run_versions = _get_versions()

    try:
        flow = cloudpickle.loads(flow_bytes)
    except Exception as exc:
        parts = ["An error occurred while unpickling the flow:", f"  {exc!r}"]
        # Check for mismatched versions to provide a better error if possible
        mismatches = []
        for name, v1 in sorted(reg_versions.items()):
            if name in run_versions:
                v2 = run_versions[name]
                if LooseVersion(v1) != v2:
                    mismatches.append(
                        f"  - {name}: (flow built with {v1!r}, currently running with {v2!r})"
                    )
        if mismatches:
            parts.append(
                "This may be due to one of the following version mismatches between "
                "the flow build and execution environments:"
            )
            parts.extend(mismatches)
        if isinstance(exc, ImportError):
            # If it's an import error, also note that the user may need to package
            # their dependencies
            prefix = "This also may" if mismatches else "This may"
            parts.append(
                f"{prefix} be due to a missing Python module in your current "
                "environment. Please ensure you have all required flow "
                "dependencies installed."
            )
        raise StorageError("\n".join(parts)) from exc

    run_prefect = run_versions["prefect"]
    reg_prefect = reg_versions.get("prefect")
    if reg_prefect and LooseVersion(reg_prefect) != run_prefect:
        # If we didn't error above, still check that the prefect versions match
        # and warn if they don't. Prefect version mismatches *may* work, but
        # they may also error later leading to confusing behavior.
        warnings.warn(
            f"This flow was built using Prefect {reg_prefect!r}, but you currently "
            f"have Prefect {run_prefect!r} installed. We recommend loading flows "
            "with the same Prefect version they were built with, failure to do so "
            "may result in errors."
        )
    return flow



from typing import Union, Iterable

import yaml

from prefect.run_configs.base import RunConfig
from prefect.utilities.filesystems import parse_path


class ECSRun(RunConfig):
    """Configure a flow-run to run as an ECS Task.

    ECS Tasks are composed of task definitions and runtime parameters.

    Task definitions can be configured using either the `task_definition`,
    `task_definition_path`, or `task_definition_arn` parameters. If neither is
    specified, the default configured on the agent will be used. At runtime
    this task definition will be registered once per flow version - subsequent
    runs of the same flow version will reuse the existing definition.

    Runtime parameters can be specified via `run_task_kwargs`. These will be
    merged with any runtime parameters configured on the agent when starting
    the task.

    Args:
        - task_definition (dict, optional): An in-memory task definition spec
            to use. See the [ECS.Client.register_task_definition][1] docs for
            more information on task definitions. Note that this definition
            will be stored directly in Prefect Cloud/Server - use
            `task_definition_path` instead if you wish to avoid this.
        - task_definition_path (str, optional): Path to a task definition spec
            to use. If a local path (no file scheme, or a `file`/`local`
            scheme), the task definition will be loaded on initialization and
            stored on the `ECSRun` object as the `task_definition` field.
            Otherwise the task definition will be loaded at runtime on the
            agent.  Supported runtime file schemes include (`s3`, `gcs`, and
            `agent` (for paths local to the runtime agent)).
        - task_definition_arn (str, optional): A pre-registered task definition
            ARN to use (either `family`, `family:version`, or a full task
            definition ARN).
        - image (str, optional): The image to use for this task. If not
            provided, will be either inferred from the flow's storage (if using
            `Docker` storage), or use the default configured on the agent.
        - env (dict, optional): Additional environment variables to set on the task.
        - cpu (int or str, optional): The amount of CPU available to the task. Can
            be ether an integer in CPU units (e.g. `1024`), or a string using
            vCPUs (e.g. `"1 vcpu"`). Note that ECS imposes strict limits on
            what values this can take, see the [ECS documentation][2] for more
            information.
        - memory (int or str, optional): The amount of memory available to the
            task. Can be ether an integer in MiB (e.g. `1024`), or a string
            with units (e.g. `"1 GB"`). Note that ECS imposes strict limits on
            what values this can take, see the [ECS documentation][2] for more
            information.
        - task_role_arn (str, optional): The name or full ARN for the IAM role
            to use for this task. If not provided, the default on the agent
            will be used (if configured).
        - execution_role_arn (str, optional): The execution role ARN to use
            when registering a task definition for this task. If not provided,
            the default on the agent will be used (if configured).
        - run_task_kwargs (dict, optional): Additional keyword arguments to
            pass to `run_task` when starting this task. See the
            [ECS.Client.run_task][3] docs for more information.
        - labels (Iterable[str], optional): An iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work

    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = ECSRun()
    ```

    Use the default task definition, but override the image and CPU:

    ```python
    flow.run_config = ECSRun(
        image="example/my-custom-image:latest",
        cpu="2 vcpu",
    )
    ```

    Use an explicit task definition stored in s3, but override the image and CPU:

    ```python
    flow.run_config = ECSRun(
        task_definition="s3://bucket/path/to/task.yaml",
        image="example/my-custom-image:latest",
        cpu="2 vcpu",
    )
    ```

    [1]: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/\
ecs.html#ECS.Client.register_task_definition

    [2]: https://docs.aws.amazon.com/AmazonECS/latest/userguide/task-cpu-memory-error.html

    [3]: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/\
ecs.html#ECS.Client.run_task
    """

    def __init__(
        self,
        *,
        task_definition: dict = None,
        task_definition_path: str = None,
        task_definition_arn: str = None,
        image: str = None,
        env: dict = None,
        cpu: Union[int, str] = None,
        memory: Union[int, str] = None,
        task_role_arn: str = None,
        execution_role_arn: str = None,
        run_task_kwargs: dict = None,
        labels: Iterable[str] = None,
    ) -> None:
        super().__init__(labels=labels)

        if (
            sum(
                [
                    task_definition is not None,
                    task_definition_path is not None,
                    task_definition_arn is not None,
                ]
            )
            > 1
        ):
            raise ValueError(
                "Can only provide one of `task_definition`, `task_definition_path`, "
                "or `task_definition_arn`"
            )
        if task_definition_path is not None:
            parsed = parse_path(task_definition_path)
            if parsed.scheme == "file":
                with open(parsed.path) as f:
                    task_definition = yaml.safe_load(f)
                    task_definition_path = None

        if cpu is not None:
            cpu = str(cpu)
        if memory is not None:
            memory = str(memory)

        self.task_definition = task_definition
        self.task_definition_path = task_definition_path
        self.task_definition_arn = task_definition_arn
        self.image = image
        self.env = env
        self.cpu = cpu
        self.memory = memory
        self.task_role_arn = task_role_arn
        self.execution_role_arn = execution_role_arn
        self.run_task_kwargs = run_task_kwargs



from typing import Iterable, Set  # noqa

import prefect


class RunConfig:
    """
    Base class for RunConfigs.

    A "run config" is an object for configuring a flow run, which maps to a
    specific agent backend.

    Args:
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work
    """

    def __init__(self, labels: Iterable[str] = None):
        self.labels = set(labels) if labels else set()  # Set[str]

    def serialize(self) -> dict:
        """
        Returns a serialized version of the RunConfig.

        Returns:
            - dict: the serialized RunConfig
        """
        schema = prefect.serialization.run_config.RunConfigSchema()
        return schema.dump(self)


class UniversalRun(RunConfig):
    """
    Configure a flow-run to run universally on any Agent.

    Unlike the other agent-specific `RunConfig` classes (e.g. `LocalRun` for
    the Local Agent), the `UniversalRun` run config is compatible with any
    agent. This can be useful for flows that don't require any custom
    configuration other than flow labels, allowing for transitioning a flow
    between agent types without any config changes.

    Args:
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work

    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = UniversalRun()
    ```

    Configure additional labels:

    ```python
    flow.run_config = UniversalRun(labels=["label-1", "label-2"])
    ```
    """



import os
import socket
from typing import TYPE_CHECKING, Any, Dict, List

import pendulum
from slugify import slugify

import prefect
from prefect.engine.results import LocalResult
from prefect.storage import Storage
from prefect.utilities.storage import (
    extract_flow_from_file,
    extract_flow_from_module,
    flow_from_bytes_pickle,
    flow_to_bytes_pickle,
)

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class Local(Storage):
    """
    Local storage class.  This class represents the Storage
    interface for Flows stored as bytes in the local filesystem.

    Note that if you register a Flow with Prefect Cloud using this storage,
    your flow will automatically be labeled with your machine's hostname. This
    ensures that only agents that are known to be running on the same
    filesystem can run your flow.

    Args:
        - directory (str, optional): the directory the flows will be stored in;
            defaults to `~/.prefect/flows`.  If it doesn't already exist, it will be
            created for you.
        - validate (bool, optional): a boolean specifying whether to validate the
            provided directory path; if `True`, the directory will be converted to an
            absolute path and created.  Defaults to `True`
        - path (str, optional): a direct path to the location of the flow file if
            `stored_as_script=True`, otherwise this path will be used when storing the serialized,
            pickled flow. If `stored_as_script=True`, the direct path may be a file path
            (such as 'path/to/myflow.py') or a direct python path (such as 'myrepo.mymodule.myflow')
        - stored_as_script (bool, optional): boolean for specifying if the flow has been stored
            as a `.py` file. Defaults to `False`
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        directory: str = None,
        validate: bool = True,
        path: str = None,
        stored_as_script: bool = False,
        **kwargs: Any,
    ) -> None:
        directory = directory or os.path.join(prefect.config.home_dir, "flows")
        self.flows = dict()  # type: Dict[str, str]
        self._flows = dict()  # type: Dict[str, "prefect.core.flow.Flow"]

        self.path = path

        if validate:
            abs_directory = os.path.abspath(os.path.expanduser(directory))
            os.makedirs(abs_directory, exist_ok=True)
        else:
            abs_directory = directory

        self.directory = abs_directory
        result = LocalResult(self.directory, validate_dir=validate)
        super().__init__(result=result, stored_as_script=stored_as_script, **kwargs)

    @property
    def default_labels(self) -> List[str]:
        if self.add_default_labels:
            return [socket.gethostname()]
        else:
            return []

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        # check if the path given is a file path
        if os.path.isfile(flow_location):
            if self.stored_as_script:
                return extract_flow_from_file(
                    file_path=flow_location, flow_name=flow_name
                )
            else:
                with open(flow_location, "rb") as f:
                    return flow_from_bytes_pickle(f.read())
        # otherwise the path is given in the module format
        else:
            return extract_flow_from_module(module_str=flow_location)

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the newly added flow in this Storage object

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        if self.stored_as_script:
            if not self.path:
                raise ValueError(
                    "A `path` must be provided to show where flow `.py` file is stored."
                )
            flow_location = self.path
        else:
            if self.path:
                flow_location = self.path
            else:
                flow_location = os.path.join(
                    self.directory,
                    slugify(flow.name),
                    slugify(pendulum.now("utc").isoformat()),
                )
            os.makedirs(os.path.dirname(flow_location), exist_ok=True)
            with open(flow_location, "wb") as f:
                f.write(flow_to_bytes_pickle(flow))

        self.flows[flow.name] = flow_location
        self._flows[flow.name] = flow
        return flow_location

    def __contains__(self, obj: Any) -> bool:
        """
        Method for determining whether an object is contained within this storage.
        """
        if not isinstance(obj, str):
            return False
        return obj in self.flows

    def build(self) -> "Storage":
        """
        Build the Storage object.

        Returns:
            - Storage: a Storage object that contains information about how and where
                each flow is stored
        """
        self.run_basic_healthchecks()
        return self



from typing import TYPE_CHECKING, Any, Dict
from urllib.parse import quote_plus

from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow


class GitLab(Storage):
    """
    GitLab storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a GitLab repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has GitLab storage:

    ```python
    flow = Flow("my-flow")
    # Can also use `repo="123456"`
    flow.storage = GitLab(repo="my/repo", path="/flows/flow.py", ref="my-branch")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register flow -f flow.py` to register this flow with GitLab storage.

    Args:
        - repo (str): the project path (i.e., 'namespace/project') or ID
        - host (str, optional): If using Gitlab server, the server host. If not specified, defaults
            to Gitlab cloud.
        - path (str, optional): a path pointing to a flow file in the repo
        - ref (str, optional): a commit SHA-1 value or branch name
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        repo: str,
        host: str = None,
        path: str = None,
        ref: str = None,
        **kwargs: Any,
    ) -> None:
        self.flows = dict()  # type: Dict[str, str]
        self._flows = dict()  # type: Dict[str, "Flow"]
        self.repo = repo
        self.host = host
        self.path = path
        self.ref = ref

        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        ref = self.ref or "master"

        from gitlab.exceptions import GitlabAuthenticationError, GitlabGetError

        try:
            project = self._gitlab_client.projects.get(quote_plus(self.repo))
            contents = project.files.get(file_path=flow_location, ref=ref)
        except GitlabAuthenticationError:
            self.logger.error(
                "Unable to authenticate Gitlab account. Please check your credentials."
            )
            raise
        except GitlabGetError:
            self.logger.error(
                f"Error retrieving file contents at {flow_location} in {self.repo}@{ref}. "
                "Ensure the project and file exist."
            )
            raise

        return extract_flow_from_file(
            file_contents=contents.decode(), flow_name=flow_name
        )

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesytem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    def build(self) -> "Storage":
        """
        Build the GitLab storage object and run basic healthchecks. Due to this object
        supporting file based storage no files are committed to the repository during
        this step. Instead, all files should be committed independently.

        Returns:
            - Storage: a GitLab object that contains information about how and where
                each flow is stored
        """
        self.run_basic_healthchecks()

        return self

    def __contains__(self, obj: Any) -> bool:
        """
        Method for determining whether an object is contained within this storage.
        """
        if not isinstance(obj, str):
            return False
        return obj in self.flows

    @property
    def _gitlab_client(self):  # type: ignore
        from prefect.utilities.git import get_gitlab_client

        return get_gitlab_client(host=self.host)



from prefect import Task
from prefect.engine.signals import FAIL
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class BatchSubmit(Task):
    """
    Task for submitting a job to AWS batch.

    For authentication, there are two options: you can set the `AWS_CREDENTIALS` Prefect Secret
    containing your AWS access keys which will be passed directly to the `boto3` client, or you
    can [configure your flow's runtime
    environment](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#guide-configuration)
    for `boto3`.

    Args:
        - job_name (str, optional): The AWS batch job name.
        - job_definition (str, optional): The AWS batch job definition.
        - job_queue (str, optional): Name of the AWS batch job queue.
        - boto_kwargs (dict, optional): additional kekyword arguments to forward to the boto client.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        job_name: str = None,
        job_definition: str = None,
        job_queue: str = None,
        boto_kwargs: dict = None,
        **kwargs,
    ):
        self.job_name = job_name
        self.job_definition = job_definition
        self.job_queue = job_queue

        if boto_kwargs is None:
            self.boto_kwargs = {}
        else:
            self.boto_kwargs = boto_kwargs

        super().__init__(**kwargs)

    @defaults_from_attrs("job_name", "job_definition", "job_queue")
    def run(
        self,
        job_name: str = None,
        job_definition: str = None,
        job_queue: str = None,
        batch_kwargs: dict = None,
        credentials: str = None,
    ):
        """
        Submit a job to the AWS Batch job service.

        Args:
            - job_name (str, optional): The AWS batch job name.
            - job_definition (str, optional): The AWS batch job definition.
            - job_queue (str, optional): Name of the AWS batch job queue.
            - batch_kwargs (dict, optional): Additional keyword arguments to pass to the boto3
                `submit_job` function. See the [submit_job](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.submit_job)  # noqa
                documentation for more details.
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.
        """
        if not job_name:
            raise ValueError("A job name must be provided.")

        if not job_definition:
            raise ValueError("A job definition must be provided.")

        if not job_queue:
            raise ValueError("A job queue must be provided.")

        if not batch_kwargs:
            batch_kwargs = {}

        batch_client = get_boto_client(
            "batch", credentials=credentials, **self.boto_kwargs
        )

        try:
            response = batch_client.submit_job(
                jobName=job_name,
                jobQueue=job_queue,
                jobDefinition=job_definition,
                **batch_kwargs,
            )
        except Exception:
            raise FAIL(f"Failed to submit job '{job_name}' to AWS Batch.")

        if not response.get("jobId"):
            raise FAIL(f"AWS Batch submit response contains no job ID: {response}")

        return response["jobId"]



"""
Tasks for interacting with the Prefect API
"""

from prefect.tasks.prefect.flow_run import FlowRunTask, StartFlowRun
from prefect.tasks.prefect.flow_run_rename import RenameFlowRunTask, RenameFlowRun
from prefect.tasks.prefect.flow_run_cancel import CancelFlowRunTask, CancelFlowRun



"""
The Prefect Storage interface encapsulates logic for storing, serializing and even running Flows.
Each storage unit is able to store _multiple_ flows (possibly with the constraint of name uniqueness
within a given unit), and exposes the following methods and attributes:

- a `flows` attribute that is a dictionary of flow name  -> location
- an `add_flow(flow: Flow) -> str` method for adding flows to Storage, and that will return the intended
location of the given flow in the Storage unit (note flow uploading/saving does not happen until `build`)
- the `__contains__(self, obj) -> bool` special method for determining whether the Storage contains a
given Flow
- one of `get_flow(flow_name: str)` for retrieving a way of interfacing with either `flow.run` or a
`FlowRunner` for the flow
- a `build() -> Storage` method for "building" the storage. In storage options where flows are stored in
an external service (such as S3 and the filesystem) the flows are uploaded/saved during this step
- a `serialize() -> dict` method for serializing the relevant information about this Storage for later
re-use.

The default flow storage mechanism is based on pickling the flow object using `cloudpickle` and the
saving that pickle to a location. Flows can optionally also be stored as a script using the
`stored_as_script` boolean kwarg. For more information visit the
[file-based storage idiom](/core/idioms/file-based.html).
"""

from warnings import warn

import prefect
from prefect import config
from .base import Storage
from .docker import Docker
from .local import Local
from .azure import Azure
from .gcs import GCS
from .s3 import S3
from .github import GitHub
from .gitlab import GitLab
from .bitbucket import Bitbucket
from .codecommit import CodeCommit
from .webhook import Webhook


def get_default_storage_class() -> type:
    """
    Returns the `Storage` class specified in
    `prefect.config.flows.defaults.storage.default_class`. If the value is a string, it will
    attempt to load the already-imported object. Otherwise, the value is returned.

    Defaults to `Local` if the string config value can not be loaded
    """
    config_value = config.flows.defaults.storage.default_class
    if isinstance(config_value, str):
        try:
            return prefect.utilities.serialization.from_qualified_name(config_value)
        except ValueError:
            warn(
                f"Could not import {config_value}; using prefect.storage.Local instead."
            )
            return Local
    else:
        return config_value



"""
This module contains a collection of tasks for interacting with Dremio Query Engine via
the pyarrow library.
"""
try:
    from prefect.tasks.dremio.dremio import DremioFetch
except ImportError as import_error:
    raise ImportError(
        'Using `prefect.tasks.dremio` requires Prefect to be installed with the "dremio" extra.'
    ) from import_error



import click

from prefect import config
from prefect.utilities.configuration import set_temporary_config
from prefect.utilities.serialization import from_qualified_name

COMMON_START_OPTIONS = [
    click.option(
        "--token",
        "-t",
        required=False,
        help="A Prefect Cloud API token with RUNNER scope.",
    ),
    click.option("--api", "-a", required=False, help="A Prefect API URL."),
    click.option(
        "--agent-config-id",
        help="An agent ID to link this agent instance with",
    ),
    click.option(
        "--name",
        "-n",
        help="A name to use for the agent",
    ),
    click.option(
        "--label",
        "-l",
        multiple=True,
        help="Labels the agent will use to query for flow runs.",
    ),
    click.option(
        "--env",
        "-e",
        multiple=True,
        help="Environment variables to set on each submitted flow run.",
    ),
    click.option(
        "--max-polls",
        help=(
            "Maximum number of times the agent should poll the Prefect API for flow "
            "runs. Default is no limit"
        ),
        type=int,
    ),
    click.option(
        "--agent-address",
        help="Address to serve internal api server at. Defaults to no server.",
        type=str,
    ),
    click.option(
        "--no-cloud-logs",
        is_flag=True,
        help="Turn off logging for all flows run through this agent.",
    ),
    click.option(
        "--log-level",
        type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
        default=None,
        help=(
            "The agent log level to use. Defaults to the value configured in your "
            "environment."
        ),
    ),
]


COMMON_INSTALL_OPTIONS = [
    click.option(
        "--token",
        "-t",
        help="A Prefect Cloud API token with RUNNER scope.",
    ),
    click.option(
        "--label",
        "-l",
        multiple=True,
        help="Labels the agent will use to query for flow runs.",
    ),
    click.option(
        "--env",
        "-e",
        multiple=True,
        help="Environment variables to set on each submitted flow run.",
    ),
]


def add_options(options):
    """A decorator for adding a list of options to a click command"""

    def decorator(func):
        for opt in reversed(options):
            func = opt(func)
        return func

    return decorator


def start_agent(agent_cls, token, api, label, env, log_level, **kwargs):
    labels = sorted(set(label))
    env_vars = dict(e.split("=", 2) for e in env)

    tmp_config = {
        "cloud.agent.auth_token": token or config.cloud.agent.auth_token,
        "cloud.agent.level": log_level or config.cloud.agent.level,
        "cloud.api": api or config.cloud.api,
    }
    with set_temporary_config(tmp_config):
        agent = agent_cls(labels=labels, env_vars=env_vars, **kwargs)
        agent.start()


@click.group()
def agent():
    """Manage Prefect agents."""


###############
# Local Agent #
###############


@agent.group()
def local():
    """Manage Prefect Local agents."""


@local.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--import-path",
    "-p",
    "import_paths",
    multiple=True,
    help="Import paths the local agent will add to all flow runs.",
)
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    is_flag=True,
)
@click.option(
    "--storage-labels/--no-storage-labels",
    default=None,
    help="Add all storage labels to the LocalAgent. DEPRECATED",
)
@click.option(
    "--hostname-label/--no-hostname-label",
    default=True,
    help="Add hostname to the LocalAgent's labels",
)
def start(import_paths, **kwargs):
    """Start a local agent"""
    from prefect.agent.local import LocalAgent

    start_agent(LocalAgent, import_paths=list(import_paths), **kwargs)


@local.command()
@add_options(COMMON_INSTALL_OPTIONS)
@click.option(
    "--import-path",
    "-p",
    "import_paths",
    multiple=True,
    help="Import paths the local agent will add to all flow runs.",
)
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    is_flag=True,
)
def install(label, env, import_paths, **kwargs):
    """Generate a supervisord.conf file for a Local agent"""
    from prefect.agent.local import LocalAgent

    conf = LocalAgent.generate_supervisor_conf(
        labels=sorted(set(label)),
        env_vars=dict(e.split("=", 2) for e in env),
        import_paths=list(import_paths),
        **kwargs,
    )
    click.echo(conf)


################
# Docker Agent #
################


@agent.group()
def docker():
    """Manage Prefect Docker agents."""


@docker.command()
@add_options(COMMON_START_OPTIONS)
@click.option("--base-url", "-b", help="Docker daemon base URL.")
@click.option("--no-pull", is_flag=True, help="Disable pulling images in the agent")
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    is_flag=True,
)
@click.option(
    "--volume",
    "volumes",
    multiple=True,
    help=(
        "Host paths for Docker bind mount volumes attached to each Flow "
        "container. Can be provided multiple times to pass multiple volumes "
        "(e.g. `--volume /volume1 --volume /volume2`)"
    ),
)
@click.option(
    "--network",
    help="Add containers to an existing docker network",
)
@click.option(
    "--no-docker-interface",
    is_flag=True,
    help=(
        "Disable the check of a Docker interface on this machine. "
        "Note: This is mostly relevant for some Docker-in-Docker "
        "setups that users may be running their agent with."
    ),
)
def start(volumes, no_docker_interface, **kwargs):
    """Start a docker agent"""
    from prefect.agent.docker import DockerAgent

    start_agent(
        DockerAgent,
        volumes=list(volumes),
        docker_interface=not no_docker_interface,
        **kwargs,
    )


####################
# Kubernetes Agent #
####################


@agent.group()
def kubernetes():
    """Manage Prefect Kubernetes agents."""


@kubernetes.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--namespace",
    help="Kubernetes namespace to deploy in. Defaults to `default`.",
)
@click.option(
    "--job-template",
    "job_template_path",
    help="Path to a kubernetes job template to use instead of the default.",
)
@click.option(
    "--service-account-name",
    "service_account_name",
    help="A default service account name to configure on started jobs.",
)
@click.option(
    "--image-pull-secrets",
    "image_pull_secrets",
    help="Default image pull secrets to configure on started jobs. Multiple "
    "values can be provided as a comma-separated list "
    "(e.g. `--image-pull-secrets VAL1,VAL2`)",
)
def start(image_pull_secrets=None, **kwargs):
    """Start a Kubernetes agent"""
    from prefect.agent.kubernetes import KubernetesAgent

    if image_pull_secrets is not None:
        image_pull_secrets = [s.strip() for s in image_pull_secrets.split(",")]

    start_agent(KubernetesAgent, image_pull_secrets=image_pull_secrets, **kwargs)


@kubernetes.command()
@add_options(COMMON_INSTALL_OPTIONS)
@click.option("--api", "-a", required=False, help="A Prefect API URL.")
@click.option("--namespace", "-n", help="Agent namespace to launch workloads.")
@click.option(
    "--image-pull-secrets",
    "-i",
    help="Name of image pull secrets to use for workloads.",
)
@click.option("--rbac", is_flag=True, help="Enable default RBAC.")
@click.option("--latest", is_flag=True, help="Use the latest Prefect image.")
@click.option("--mem-request", help="Requested memory for Prefect init job.")
@click.option("--mem-limit", help="Limit memory for Prefect init job.")
@click.option("--cpu-request", help="Requested CPU for Prefect init job.")
@click.option("--cpu-limit", help="Limit CPU for Prefect init job.")
@click.option("--image-pull-policy", help="imagePullPolicy for Prefect init job")
@click.option(
    "--service-account-name", help="Name of Service Account for Prefect init job"
)
@click.option("--backend", "-b", help="Prefect backend to use for this agent.")
def install(label, env, **kwargs):
    """Generate a supervisord.conf file for a Local agent"""
    from prefect.agent.kubernetes import KubernetesAgent

    deployment = KubernetesAgent.generate_deployment_yaml(
        labels=sorted(set(label)), env_vars=dict(e.split("=", 2) for e in env), **kwargs
    )
    click.echo(deployment)


#################
# Fargate Agent #
#################


def warn_fargate_deprecated():
    click.secho(
        "Warning: The Fargate agent is deprecated, please transition to using the ECS agent instead",
        fg="yellow",
        err=True,
    )


@agent.group()
def fargate():
    """Manage Prefect Fargate agents (DEPRECATED).

    The Fargate agent is deprecated, please transition to using the ECS agent instead.
    """


@fargate.command(
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True)
)
@add_options(COMMON_START_OPTIONS)
@click.pass_context
def start(ctx, **kwargs):
    """Start a Fargate agent (DEPRECATED)

    The Fargate agent is deprecated, please transition to using the ECS agent instead.
    """
    from prefect.agent.fargate import FargateAgent

    warn_fargate_deprecated()

    for item in ctx.args:
        k, v = item.replace("--", "").split("=", 2)
        kwargs[k] = v

    start_agent(FargateAgent, _called_from_cli=True, **kwargs)


#############
# ECS Agent #
#############


@agent.group()
def ecs():
    """Manage Prefect ECS agents."""


@ecs.command()
@add_options(COMMON_START_OPTIONS)
@click.option(
    "--cluster",
    help="The cluster to use. If not provided, your default cluster will be used",
)
@click.option(
    "--launch-type",
    type=click.Choice(["FARGATE", "EC2"], case_sensitive=False),
    help="The launch type to use, defaults to FARGATE",
)
@click.option(
    "--task-role-arn",
    help="The default task role ARN to use for ECS tasks started by this agent.",
)
@click.option(
    "--execution-role-arn",
    help="The default execution role ARN to use for ECS tasks started by this agent.",
)
@click.option(
    "--task-definition",
    "task_definition_path",
    help=(
        "Path to a task definition template to use when defining new tasks "
        "instead of the default."
    ),
)
@click.option(
    "--run-task-kwargs",
    "run_task_kwargs_path",
    help="Path to a yaml file containing extra kwargs to pass to `run_task`",
)
def start(**kwargs):
    """Start an ECS agent"""
    from prefect.agent.ecs import ECSAgent

    start_agent(ECSAgent, **kwargs)


#############################
# Deprecated Agent Commands #
#############################

_agents = {
    "fargate": "prefect.agent.fargate.FargateAgent",
    "docker": "prefect.agent.docker.DockerAgent",
    "kubernetes": "prefect.agent.kubernetes.KubernetesAgent",
    "local": "prefect.agent.local.LocalAgent",
}


@agent.command(
    hidden=True,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.argument("agent-option", default="local")
@click.option(
    "--token", "-t", required=False, help="A Prefect Cloud API token.", hidden=True
)
@click.option("--api", "-a", required=False, help="A Prefect API URL.", hidden=True)
@click.option("--agent-config-id", required=False, help="An agent ID", hidden=True)
@click.option(
    "--name",
    "-n",
    required=False,
    help="A name to use for the agent",
    hidden=True,
    default=None,
)
@click.option(
    "--verbose", "-v", is_flag=True, help="Enable verbose agent logs.", hidden=True
)
@click.option(
    "--label",
    "-l",
    multiple=True,
    help="Labels the agent will use to query for flow runs.",
    hidden=True,
)
@click.option(
    "--storage-labels/--no-storage-labels",
    default=None,
    help="Add all storage labels to the LocalAgent. DEPRECATED",
)
@click.option(
    "--hostname-label/--no-hostname-label",
    default=True,
    help="Add hostname to the LocalAgent's labels",
    hidden=True,
)
@click.option(
    "--env",
    "-e",
    multiple=True,
    help="Environment variables to set on each submitted flow run.",
    hidden=True,
)
@click.option(
    "--max-polls",
    required=False,
    help="Maximum number of polls for the agent",
    hidden=True,
    type=int,
)
@click.option(
    "--agent-address",
    required=False,
    help="Address to serve internal api server at. Defaults to no server.",
    hidden=True,
    type=str,
    default="",
)
@click.option(
    "--namespace",
    required=False,
    help="Kubernetes namespace to create jobs.",
    hidden=True,
)
@click.option(
    "--job-template",
    required=False,
    help="Path to a kubernetes job template to use instead of the default.",
    hidden=True,
)
@click.option(
    "--import-path",
    "-p",
    multiple=True,
    help="Import paths the local agent will add to all flow runs.",
    hidden=True,
)
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    hidden=True,
    is_flag=True,
)
@click.option("--no-pull", is_flag=True, help="Pull images flag.", hidden=True)
@click.option(
    "--no-cloud-logs",
    is_flag=True,
    help="Turn off logging for all flows run through this agent.",
    hidden=True,
)
@click.option("--base-url", "-b", help="Docker daemon base URL.", hidden=True)
@click.option(
    "--volume",
    multiple=True,
    help="Host paths for Docker bind mount volumes attached to each Flow runtime container.",
    hidden=True,
)
@click.option(
    "--network",
    help="Add containers to an existing docker network",
    hidden=True,
)
@click.option(
    "--no-docker-interface",
    is_flag=True,
    help="Disable presence of a Docker interface.",
    hidden=True,
)
@click.pass_context
def start(
    ctx,
    agent_option,
    token,
    api,
    agent_config_id,
    name,
    verbose,
    label,
    env,
    namespace,
    job_template,
    no_pull,
    no_cloud_logs,
    base_url,
    import_path,
    show_flow_logs,
    volume,
    network,
    no_docker_interface,
    max_polls,
    agent_address,
    hostname_label,
    storage_labels,
):
    """
    Start an agent.

    DEPRECATED: use `prefect agent <agent-type> start` instead.

    \b
    Arguments:
        agent-option    TEXT    The name of an agent to start (e.g. `docker`, `kubernetes`,
                                `local`, `fargate`). Defaults to `local`

    \b
    Options:
        --token, -t             TEXT    A Prefect Cloud API token with RUNNER scope
        --api, -a               TEXT    A Prefect API URL
        --agent-config--id      TEXT    An agent ID to link this agent instance with
        --name, -n              TEXT    A name to use for the agent
        --verbose, -v                   Enable verbose agent DEBUG logs
                                        Defaults to INFO level logging
        --label, -l             TEXT    Labels the agent will use to query for flow runs
                                        Multiple values supported e.g. `-l label1 -l label2`
        --env, -e               TEXT    Environment variables to set on each submitted flow
                                        run.
                                        Note that equal signs in environment variable values
                                        are not currently supported from the CLI.  Multiple
                                        values supported.
                                            e.g. `-e AUTH=token -e PKG_SETTING=true`
        --max-polls             INT     Maximum number of times the agent should poll the
                                        Prefect API for flow runs. Will run forever if not
                                        specified.
        --no-cloud-logs                 Turn off logging to the Prefect API for all flow runs
                                        Defaults to `False`
        --agent-address         TEXT    The address to server internal api at. Currently this
                                        is just health checks for use by an orchestration layer
                                        (e.g. kubernetes). Leave blank for no api server (default).

    \b
    Local Agent:
        --import-path, -p   TEXT    Import paths which will be provided to each Flow's
                                    runtime environment.  Used for Flows which might
                                    import from scripts or local packages.  Multiple values
                                    supported.
                                        e.g. `-p /root/my_scripts -p /utilities`
        --show-flow-logs, -f        Display logging output from flows run by the agent
                                    (available for Local and Docker agents only)
        --hostname-label            Add hostname to the Agent's labels
                                        (Default to True. Disable with --no-hostname-label option)

    \b
    Docker Agent:
        --base-url, -b      TEXT    A Docker daemon host URL for a DockerAgent
        --no-pull                   Pull images for a DockerAgent
                                    Defaults to pulling if not provided
        --volume            TEXT    Host paths for Docker bind mount volumes attached to
                                    each Flow runtime container. Multiple values supported.
                                        e.g. `--volume /some/path`
        --network           TEXT    Add containers to an existing docker network
        --no-docker-interface       Disable the check of a Docker interface on this machine.
                                    Note: This is mostly relevant for some Docker-in-Docker
                                    setups that users may be running their agent with.

    \b
    Kubernetes Agent:
        --namespace     TEXT    A Kubernetes namespace to create Prefect jobs in
                                Defaults to env var `NAMESPACE` or `default`
        --job-template  TEXT    Path to a job template to use instead of the default.

    \b
    Fargate Agent Options:
        Any of the configuration options outlined in the docs can be provided here
        https://docs.prefect.io/orchestration/agents/fargate.html#configuration
    """
    # Split context
    kwargs = dict()
    for item in ctx.args:
        item = item.replace("--", "")
        kwargs.update([item.split("=")])

    tmp_config = {
        "cloud.agent.auth_token": token or config.cloud.agent.auth_token,
    }
    if verbose:
        tmp_config["cloud.agent.level"] = "DEBUG"
    if api:
        tmp_config["cloud.api"] = api

    with set_temporary_config(tmp_config):
        retrieved_agent = _agents.get(agent_option)

        if not retrieved_agent:
            click.secho(
                "{} is not a valid agent".format(agent_option), fg="red", err=True
            )
            return

        if agent_option == "fargate":
            warn_fargate_deprecated()
        else:
            click.secho(
                f"Warning: `prefect agent start {agent_option}` is deprecated, use "
                f"`prefect agent {agent_option} start` instead",
                fg="yellow",
                err=True,
            )

        env_vars = dict()
        for env_var in env:
            k, v = env_var.split("=")
            env_vars[k] = v

        labels = sorted(set(label))

        if agent_option == "local":
            from_qualified_name(retrieved_agent)(
                agent_config_id=agent_config_id,
                name=name,
                labels=labels,
                env_vars=env_vars,
                max_polls=max_polls,
                agent_address=agent_address,
                import_paths=list(import_path),
                show_flow_logs=show_flow_logs,
                no_cloud_logs=no_cloud_logs,
                hostname_label=hostname_label,
                storage_labels=storage_labels,
            ).start()
        elif agent_option == "docker":
            from_qualified_name(retrieved_agent)(
                agent_config_id=agent_config_id,
                name=name,
                labels=labels,
                env_vars=env_vars,
                max_polls=max_polls,
                no_cloud_logs=no_cloud_logs,
                agent_address=agent_address,
                base_url=base_url,
                no_pull=no_pull,
                show_flow_logs=show_flow_logs,
                volumes=list(volume),
                network=network,
                docker_interface=not no_docker_interface,
            ).start()
        elif agent_option == "fargate":
            from_qualified_name(retrieved_agent)(
                agent_config_id=agent_config_id,
                name=name,
                labels=labels,
                env_vars=env_vars,
                max_polls=max_polls,
                no_cloud_logs=no_cloud_logs,
                agent_address=agent_address,
                _called_from_cli=True,
                **kwargs,
            ).start()
        elif agent_option == "kubernetes":
            from_qualified_name(retrieved_agent)(
                agent_config_id=agent_config_id,
                namespace=namespace,
                job_template_path=job_template,
                name=name,
                labels=labels,
                env_vars=env_vars,
                max_polls=max_polls,
                no_cloud_logs=no_cloud_logs,
                agent_address=agent_address,
            ).start()
        else:
            from_qualified_name(retrieved_agent)(
                agent_config_id=agent_config_id,
                name=name,
                labels=labels,
                env_vars=env_vars,
                max_polls=max_polls,
                no_cloud_logs=no_cloud_logs,
                agent_address=agent_address,
            ).start()


@agent.command(hidden=True)
@click.argument("name")
@click.option(
    "--token", "-t", required=False, help="A Prefect Cloud API token.", hidden=True
)
@click.option("--api", "-a", required=False, help="A Prefect API URL.", hidden=True)
@click.option(
    "--namespace",
    "-n",
    required=False,
    help="Agent namespace to launch workloads.",
    hidden=True,
)
@click.option(
    "--image-pull-secrets",
    "-i",
    required=False,
    help="Name of image pull secrets to use for workloads.",
    hidden=True,
)
@click.option("--rbac", is_flag=True, help="Enable default RBAC.", hidden=True)
@click.option(
    "--latest", is_flag=True, help="Use the latest Prefect image.", hidden=True
)
@click.option(
    "--mem-request",
    required=False,
    help="Requested memory for Prefect init job.",
    hidden=True,
)
@click.option(
    "--mem-limit",
    required=False,
    help="Limit memory for Prefect init job.",
    hidden=True,
)
@click.option(
    "--cpu-request",
    required=False,
    help="Requested CPU for Prefect init job.",
    hidden=True,
)
@click.option(
    "--cpu-limit", required=False, help="Limit CPU for Prefect init job.", hidden=True
)
@click.option(
    "--image-pull-policy",
    required=False,
    help="imagePullPolicy for Prefect init job",
    hidden=True,
)
@click.option(
    "--service-account-name",
    required=False,
    help="Name of Service Account for Prefect init job",
    hidden=True,
)
@click.option(
    "--label",
    "-l",
    multiple=True,
    help="Labels the agent will use to query for flow runs.",
    hidden=True,
)
@click.option(
    "--env",
    "-e",
    multiple=True,
    help="Environment variables to set on each submitted flow run.",
    hidden=True,
)
@click.option(
    "--import-path",
    "-p",
    multiple=True,
    help="Import paths the local agent will add to all flow runs.",
    hidden=True,
)
@click.option(
    "--show-flow-logs",
    "-f",
    help="Display logging output from flows run by the agent.",
    hidden=True,
    is_flag=True,
)
@click.option(
    "--backend",
    "-b",
    required=False,
    help="Prefect backend to use for this agent.",
    hidden=True,
)
def install(
    name,
    token,
    api,
    namespace,
    image_pull_secrets,
    rbac,
    latest,
    mem_request,
    mem_limit,
    cpu_request,
    cpu_limit,
    image_pull_policy,
    service_account_name,
    label,
    env,
    import_path,
    show_flow_logs,
    backend,
):
    """
    Install an agent. Outputs configuration text which can be used to install on various
    platforms. The Prefect image version will default to your local `prefect.__version__`

    DEPRECATED: use `prefect agent <agent-type> install` instead.

    \b
    Arguments:
        name                        TEXT    The name of an agent to install (e.g.
                                            `kubernetes`, `local`)

    \b
    Options:
        --token, -t                 TEXT    A Prefect Cloud API token
        --label, -l                 TEXT    Labels the agent will use to query for flow runs
                                            Multiple values supported.
                                                e.g. `-l label1 -l label2`
        --env, -e                   TEXT    Environment variables to set on each submitted
                                            flow run. Note that equal signs in environment
                                            variable values are not currently supported from
                                            the CLI. Multiple values supported.
                                                e.g. `-e AUTH=token -e PKG_SETTING=true`

    \b
    Kubernetes Agent:
        --api, -a                   TEXT    A Prefect API URL
        --namespace, -n             TEXT    Agent namespace to launch workloads
        --image-pull-secrets, -i    TEXT    Name of image pull secrets to use for workloads
        --rbac                              Enable default RBAC on install
        --latest                            Use the `latest` Prefect image
        --mem-request               TEXT    Requested memory for Prefect init job
        --mem-limit                 TEXT    Limit memory for Prefect init job
        --cpu-request               TEXT    Requested CPU for Prefect init job
        --cpu-limit                 TEXT    Limit CPU for Prefect init job
        --image-pull-policy         TEXT    imagePullPolicy for Prefect init job
        --service-account-name      TEXT    Name of Service Account for Prefect init job
        --backend                   TEST    Prefect backend to use for this agent
                                            Defaults to the backend currently set in config.

    \b
    Local Agent:
        --import-path, -p           TEXT    Absolute import paths to provide to the local
                                            agent. Multiple values supported.
                                                e.g. `-p /root/my_scripts -p /utilities`
        --show-flow-logs, -f                Display logging output from flows run by the
                                            agent
    """

    supported_agents = {
        "kubernetes": "prefect.agent.kubernetes.KubernetesAgent",
        "local": "prefect.agent.local.LocalAgent",
    }

    retrieved_agent = supported_agents.get(name)

    if not retrieved_agent:
        click.secho(
            "{} is not a supported agent for `install`".format(name), fg="red", err=True
        )
        return

    click.secho(
        f"Warning: `prefect agent install {name}` is deprecated, use "
        f"`prefect agent {name} install` instead",
        fg="yellow",
        err=True,
    )

    env_vars = dict()
    for env_var in env:
        k, v = env_var.split("=")
        env_vars[k] = v

    labels = sorted(set(label))
    if name == "kubernetes":
        deployment = from_qualified_name(retrieved_agent).generate_deployment_yaml(
            token=token,
            api=api,
            namespace=namespace,
            image_pull_secrets=image_pull_secrets,
            rbac=rbac,
            latest=latest,
            mem_request=mem_request,
            mem_limit=mem_limit,
            cpu_request=cpu_request,
            cpu_limit=cpu_limit,
            image_pull_policy=image_pull_policy,
            service_account_name=service_account_name,
            labels=labels,
            env_vars=env_vars,
            backend=backend,
        )
        click.echo(deployment)
    elif name == "local":
        conf = from_qualified_name(retrieved_agent).generate_supervisor_conf(
            token=token,
            labels=labels,
            env_vars=env_vars,
            import_paths=list(import_path),
            show_flow_logs=show_flow_logs,
        )
        click.echo(conf)
